import abc

import numpy as np
import pandas as pd
from numpy.random import default_rng

from badgers.core.base import GeneratorMixin
from badgers.core.decorators.tabular_data import preprocess_inputs
from badgers.core.utils import normalize_proba


class ImbalanceGenerator(GeneratorMixin):
    """
    Base class for transformers that makes tabular data imbalanced
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """
        :param random_generator: A random generator
        """
        self.random_generator = random_generator

    @abc.abstractmethod
    def generate(self, X, y=None, **params):
        pass


class RandomSamplingFeaturesGenerator(ImbalanceGenerator):

    def __init__(self, random_generator=default_rng(seed=0), ):
        """
        :param random_generator: A random generator
        """
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y=None, sampling_proba_func=lambda X: normalize_proba(X.iloc[:, 0])):
        """
        Randomly samples instances based on the features values in X

        :param X:
        :param y:
        :param sampling_proba_func: A function that takes as input data and returns a sampling probability
        :return: Xt, yt
        """
        # total number of instances that will be missing
        # sampling
        sampling_proba = sampling_proba_func(X)
        sampling_mask = self.random_generator.choice(X.shape[0], p=sampling_proba, size=X.shape[0], replace=True)
        Xt = X.iloc[sampling_mask,:]
        yt = y[sampling_mask] if y is not None else y
        return Xt, yt


class RandomSamplingClassesGenerator(ImbalanceGenerator):
    """
    Randomly samples data points within predefined classes
    """

    def __init__(self, random_generator=default_rng(seed=0), ):
        """

        :param random_generator: A random generator

        """
        super().__init__(random_generator=random_generator)
        self.transformed_labels_ = None

    @preprocess_inputs
    def generate(self, X, y, proportion_classes: dict = None):
        """
        Randomly samples instances for each classes

        :param X:
        :param y:
        :param proportion_classes: Example for having in total 50% of class 'A', 30% of class 'B', and 20% of class 'C'
            proportion_classes={'A':0.5, 'B':0.3, 'C':0.2}
        :return:
        """
        # local variables
        Xt = []
        transformed_labels = []

        for label, prop in proportion_classes.items():
            size = int(prop * X.shape[0])
            Xt.append(self.random_generator.choice(X[y == label], size=size, replace=True))
            transformed_labels += [label] * size

        Xt = pd.DataFrame(
            data=np.vstack(Xt),
            columns=X.columns
        )

        yt = pd.Series(data=transformed_labels)

        return Xt, yt


class RandomSamplingTargetsGenerator(ImbalanceGenerator):
    """
    Randomly samples data points
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """

        :param random_generator: A random generator
        :param sampling_proba_func: A function that takes y as input and returns a sampling probability
        """
        super().__init__(random_generator=random_generator)
        self.transformed_labels_ = None

    @preprocess_inputs
    def generate(self, X, y, sampling_proba_func=lambda y: normalize_proba(y)):
        """
        Randomly samples instances for each classes

        :param X:
        :param y:
        :return:
        """
        sampling_probabilities_ = sampling_proba_func(y)
        sampling_mask = self.random_generator.choice(X.shape[0], p=sampling_probabilities_, size=X.shape[0],
                                                     replace=True)

        Xt = X.iloc[sampling_mask, :]
        yt = y[sampling_mask]

        return Xt, yt



import unittest

from numpy.random import default_rng
from sklearn.datasets import make_blobs

from badgers.core.pipeline import Pipeline
from badgers.generators.tabular_data.imbalance import RandomSamplingClassesGenerator
from badgers.generators.tabular_data.noise import GaussianNoiseGenerator


class MyTestCase(unittest.TestCase):
    def test_generate(self):
        random_generator = default_rng(0)
        X, y = make_blobs(centers=3, random_state=0)
        generators = {
            'imbalance': RandomSamplingClassesGenerator(random_generator=random_generator),
            'noise': GaussianNoiseGenerator(random_generator=random_generator)
        }
        pipeline = Pipeline(generators=generators)
        params = {
            'imbalance': {'proportion_classes': {0: 0.5, 1: 0.25, 2: 0.25}},
            'noise': {'noise_std':0.5}
        }
        Xt, yt = pipeline.generate(X=X, y=y, params=params)


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest import TestCase

import numpy as np
from numpy.random import default_rng

from badgers.generators.tabular_data.noise import GaussianNoiseGenerator, GaussianNoiseClassesGenerator
from tests.generators.tabular_data import generate_test_data_with_classification_labels, \
    generate_test_data_only_features


class TestGaussianNoiseGenerator(TestCase):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.input_test_data = generate_test_data_only_features(rng=self.rng)

    def test_generate(self):
        """
        - checks that the transformed array has the same size as the input array
        - checks that the variance of the transformed array is greater than the one of the input array
        """
        generator = GaussianNoiseGenerator()
        for input_type, (X, y) in self.input_test_data.items():
            with self.subTest(transformer=generator.__class__, input_type=input_type):
                Xt, _ = generator.generate(X.copy(), y=None, noise_std=1)
                # assert arrays have same size
                self.assertEqual(len(X), len(Xt))
                self.assertTrue((np.var(Xt, axis=0) > np.var(X, axis=0)).all())


class TestGaussianNoiseClassesGenerator(TestCase):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.input_test_data = generate_test_data_with_classification_labels(rng=self.rng)

    def test_generate(self):
        """
        - checks that the transformed array has the same size as the input array
        - checks that the variance of the transformed array is greater than the one of the input array
        """

        for input_type, (X, y) in self.input_test_data.items():
            noise_std_per_class = {label: 0.1 for label in np.unique(y)}
            generator = GaussianNoiseClassesGenerator()
            with self.subTest(transformer=generator.__class__, input_type=input_type):
                Xt, yt = generator.generate(X.copy(), y, noise_std_per_class=noise_std_per_class)
                # assert arrays have same size
                self.assertEqual(len(X), len(Xt))
                self.assertEqual(len(y), len(yt))


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest import TestCase

import numpy as np
from numpy.random import default_rng

from badgers.generators.tabular_data.missingness import MissingValueGenerator
from tests.generators.tabular_data import generate_test_data_only_features


class TestMissingValueGenerator(TestCase):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.generators_classes = MissingValueGenerator.__subclasses__()
        self.input_test_data = generate_test_data_only_features(rng=self.rng)

    def test_all_generators(self):
        percentage_missing = 0.1
        for cls in self.generators_classes:
            transformer = cls()
            for input_type, (X, y) in self.input_test_data.items():
                with self.subTest(transformer=transformer.__class__, input_type=input_type):
                    Xt, _ = transformer.generate(X.copy(), y, percentage_missing=percentage_missing)
                    # assert arrays have same size
                    # assert arrays have same size
                    if input_type[-2:] == '1D':
                        self.assertEqual(Xt.shape[1], 1)
                    else:
                        self.assertEqual(Xt.shape[1], 10)

                    # assert that the right number of nans have been generated
                    if input_type[-2:] == '1D':
                        self.assertEqual(
                            np.sum(np.isnan(Xt).sum()),
                            int(percentage_missing * 100)
                        )
                    else:
                        self.assertEqual(
                            np.sum(np.isnan(Xt).sum()),
                            int(percentage_missing * 1000)
                        )



if __name__ == '__main__':
    unittest.main()



import abc
from typing import Tuple

import pandas as pd
from badgers.generators.time_series.utils import generate_random_patterns_indices
from numpy.random import default_rng
from sklearn.preprocessing import StandardScaler

from badgers.core.base import GeneratorMixin
from badgers.core.decorators.time_series import preprocess_inputs


class NoiseGenerator(GeneratorMixin):
    """
    Base class for transformers that add noise to tabular data
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """
        :param random_generator: A random generator
        """
        self.random_generator = random_generator

    @abc.abstractmethod
    def generate(self, X, y, **params) -> Tuple:
        pass


class LocalGaussianNoiseGenerator(NoiseGenerator):

    def __init__(self, random_generator=default_rng(seed=0), ):
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y, n_patterns: int = 10, min_width_pattern: int = 5, max_width_patterns: int=10,
                 noise_std: float = 0.1) -> Tuple:
        # generate extreme values indices and values
        self.patterns_indices_ = generate_random_patterns_indices(
            random_generator=self.random_generator,
            n_patterns=n_patterns,
            signal_size=len(X),
            min_width_pattern=min_width_pattern,
            max_width_patterns=max_width_patterns)

        scaler = StandardScaler()
        # fit, transform
        scaler.fit(X)
        Xt = scaler.transform(X)

        for (start, end) in self.patterns_indices_:
            Xt[start:end, :] += self.random_generator.normal(loc=0, scale=noise_std, size=(end-start, Xt.shape[1]))

        # inverse standardization
        return pd.DataFrame(data=scaler.inverse_transform(Xt), columns=X.columns, index=X.index), y


class GlobalGaussianNoiseGenerator(NoiseGenerator):
    def __init__(self, random_generator=default_rng(seed=0)):
        """
        :param random_generator: A random generator
        """
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y, noise_std: float = 0.1):
        """
        Add Gaussian white noise to the data.
        the data is first standardized (each column has a mean = 0 and variance = 1).
        The noise is generated from a normal distribution with standard deviation = `noise_std`.
        The noise is added to the data.

        :param noise_std: The standard deviation of the noise to be added
        :param X:
        :return:
        """
        scaler = StandardScaler()
        # fit, transform
        scaler.fit(X)
        Xt = scaler.transform(X)
        # add noise
        Xt = Xt + self.random_generator.normal(loc=0, scale=noise_std, size=Xt.shape)
        # inverse standardization
        return pd.DataFrame(data=scaler.inverse_transform(Xt), columns=X.columns, index=X.index), y



import unittest
from unittest import TestCase

import numpy as np
import pandas as pd

from badgers.generators.time_series.patterns import Pattern, add_offset, \
    add_linear_trend, scale, RandomlySpacedConstantPatterns, RandomlySpacedLinearPatterns, RandomlySpacedPatterns


class TestPattern(TestCase):

    def test_resample(self):
        pattern = Pattern(np.array([1, 2, 3, 4, 5]))
        resampled = pattern.resample(10)
        self.assertEqual(len(resampled), 10)
        self.assertAlmostEqual(resampled[0], 1)
        self.assertAlmostEqual(resampled[-1], 5)

    def test_add_offset(self):
        pattern = Pattern(np.array([1, 2, 3, 4, 5]))
        expected_pattern = np.array([3, 4, 5, 6, 7]).reshape(-1, 1)
        transformed_pattern = add_offset(pattern.values, 2)
        self.assertListEqual(transformed_pattern.tolist(), expected_pattern.tolist())

    def test_add_linear_trend(self):
        pattern = Pattern(np.array([1, 2, 3, 4, 5]))
        transformed_pattern = add_linear_trend(pattern.values, start_value=0, end_value=-1)
        self.assertEqual(len(transformed_pattern), 5)
        self.assertEqual(transformed_pattern[0], 0)
        self.assertEqual(transformed_pattern[-1], -1)

    def test_scale(self):
        pattern = Pattern(np.array([1, 2, 3, 4, 5]))
        expected_pattern = np.array([2, 4, 6, 8, 10]).reshape(-1, 1)
        transformed_pattern = scale(pattern.values, scaling_factor=2)
        self.assertListEqual(transformed_pattern.tolist(), expected_pattern.tolist())


class TestRandomlySpacedPatterns(TestCase):

    def test_generate_1D(self):
        n_patterns = 3
        X = np.zeros(shape=100).reshape(-1,1)
        pattern = Pattern(values=np.array([0, 1, 2, 3, 4, 5, 4, 3, 2, 1, 0]))
        generator = RandomlySpacedPatterns()
        Xt, _ = generator.generate(X=X, y=None, n_patterns=n_patterns, min_width_pattern=5, max_width_patterns=10, pattern=pattern)
        # assert that the correct number of patterns have been injected
        self.assertEqual(n_patterns, len(generator.patterns_indices_))
        # assert that no pattern overlap
        for i in range(n_patterns - 1):
            self.assertLess(generator.patterns_indices_[i][1], generator.patterns_indices_[i+1][0])


    def test_generate_2D(self):
        n_patterns = 3
        X = np.zeros(shape=(100, 2))
        pattern = Pattern(
            values=np.array([[0, 1, 2, 3, 4, 5, 4, 3, 2, 1, 0], [0, -1, -2, -3, -4, -5, -4, -3, -2, -1, 0]]).T)
        generator = RandomlySpacedPatterns()
        Xt, _ = generator.generate(X=X, y=None, n_patterns=n_patterns, min_width_pattern=5, max_width_patterns=10, pattern=pattern)
        # assert that the correct number of patterns have been injected
        self.assertEqual(n_patterns, len(generator.patterns_indices_))
        # assert that no pattern overlap
        for i in range(n_patterns - 1):
            self.assertLess(generator.patterns_indices_[i][1], generator.patterns_indices_[i+1][0])


class TestRandomlySpacedConstantPatterns(TestCase):

    def setUp(self) -> None:
        t = np.linspace(1, 10, 101)
        self.X = pd.DataFrame(data=(np.sin(t * 2 * np.pi) + 0.5).reshape(-1, 1))

    def test_generate(self):
        n_patterns = 3
        generator = RandomlySpacedConstantPatterns()
        Xt, _ = generator.generate(X=self.X, y=None, n_patterns=n_patterns, min_width_pattern=5, max_width_patterns=10,
                                                   constant_value=0)
        # assert that the correct number of patterns have been injected
        self.assertEqual(n_patterns, len(generator.patterns_indices_))
        # assert that all patterns are constant = 0
        for start, end in generator.patterns_indices_:
            self.assertListEqual(Xt.iloc[start:end, :].values.tolist(), np.zeros((end - start, self.X.shape[1])).tolist())


class TestRandomlySpacedLinearPatterns(TestCase):

    def setUp(self) -> None:
        t = np.linspace(1, 10, 101)
        self.X = pd.DataFrame(data=(np.sin(t * 2 * np.pi) + 0.5).reshape(-1, 1))

    def test_generate(self):
        generator = RandomlySpacedLinearPatterns()
        Xt, _ = generator.generate(X=self.X, y=None, n_patterns=3, min_width_pattern=5, max_width_patterns=10)
        for (start, end) in generator.patterns_indices_:
            for col in range(self.X.shape[1]):
                self.assertListEqual(Xt.iloc[start:end, col].tolist(),
                                     np.linspace(self.X.iloc[start, col], self.X.iloc[end, col], end - start).tolist())


if __name__ == '__main__':
    unittest.main()


import abc
from typing import Tuple

import numpy as np
from numpy.random import default_rng

from badgers.core.base import GeneratorMixin
from badgers.core.decorators.time_series import preprocess_inputs


class TrendsGenerator(GeneratorMixin):
    """
    Base class for transformers that generate trends in time-series data
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """
        :param random_generator: a random number generator
        """
        self.random_generator = random_generator

    @abc.abstractmethod
    def generate(self, X, y, **params) -> Tuple:
        pass


class GlobalAdditiveLinearTrendGenerator(TrendsGenerator):
    """
    Add a linear trend to the input time-series data
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y, slope) -> Tuple:
        """

        :param X:
        :param y:
        :param slope:
        :return:
        """

        t = np.linspace(0, 1, len(X))
        trend = t[:, np.newaxis] * slope
        Xt = X.add(trend, axis=0)
        return Xt, y



import unittest
from unittest import TestCase

import numpy as np
from numpy.random import default_rng

from badgers.core.utils import normalize_proba, random_sign, random_spherical_coordinate


class TestUtils(TestCase):

    def setUp(self) -> None:
        self.rng = default_rng(0)

    def test_normalize_proba(self):
        p = self.rng.uniform(0, 1, size=(3, 5))
        p = normalize_proba(p)
        self.assertEqual(p.shape, (3, 5))
        for s in p.sum(axis=0).tolist():
            self.assertAlmostEqual(s, 1.0)

    def test_random_signs(self):
        for X in [np.ones(1), np.ones(10), np.ones((10, 10))]:
            Xt = random_sign(random_generator=self.rng, size=X.shape)
            with self.subTest(X=X):
                self.assertEqual(
                    X.shape, Xt.shape
                )
                self.assertTrue(
                    np.all(X == np.abs(Xt))
                )
                if X.shape != (1,):
                    self.assertNotEqual(
                        np.sum(X == 1), np.sum(Xt == 1)
                    )

    def test_random_spherical_coordinate(self):
        size = 2
        r = 1
        x = random_spherical_coordinate(random_generator=self.rng, size=size, radius=r)
        self.assertAlmostEqual(np.sum(x ** 2), r ** 2)


if __name__ == '__main__':
    unittest.main()



from unittest import TestCase

import numpy as np
from numpy.random import default_rng

from badgers.generators.tabular_data.drift import RandomShiftGenerator, RandomShiftClassesGenerator
from tests.generators.tabular_data import generate_test_data_with_classification_labels, \
    generate_test_data_only_features


class TestRandomShiftGenerator(TestCase):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.test_data = generate_test_data_only_features(rng=self.rng)

    def test_generate_shift_std_number(self):
        """
        Asserts:
        - Arrays X and Xt have the same shape
        """
        generator = RandomShiftGenerator(random_generator=self.rng)
        for input_type, (X, y) in self.test_data.items():
            with self.subTest(transformer=generator.__class__, input_type=input_type):
                Xt, _ = generator.generate(X.copy(), y, shift_std=0.1)
                # assert arrays have same size
                self.assertEqual(len(X), len(Xt))

    def test_generate_shift_std_array(self):
        """
        Asserts:
        - Arrays X and Xt have the same shape
        """
        generator = RandomShiftGenerator(random_generator=self.rng)
        for input_type, (X, y) in self.test_data.items():
            if input_type[-2:] == '1D':
                shift_std = np.array([0.1])
            else:
                shift_std = np.linspace(0.1, 1, 10)
            with self.subTest(transformer=generator.__class__, input_type=input_type):
                Xt, _ = generator.generate(X.copy(), y, shift_std=shift_std)
                # assert arrays have same size
                self.assertEqual(len(X), len(Xt))


class TestRandomShiftGenerator(TestCase):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.test_data = generate_test_data_with_classification_labels(rng=self.rng)

    def test_generate_shift_std_number(self):
        """
        Asserts:
        - Arrays X and Xt have the same shape
        """
        generator = RandomShiftClassesGenerator(random_generator=self.rng)
        for input_type, (X, y) in self.test_data.items():
            with self.subTest(transformer=generator.__class__, input_type=input_type):
                Xt, yt = generator.generate(X.copy(), y, shift_std=0.1)
                # assert arrays have same size
                self.assertEqual(len(X), len(Xt))

    def test_generate_shift_std_array(self):
        """
        Asserts:
        - Arrays X and Xt have the same shape
        """
        generator = RandomShiftClassesGenerator(random_generator=self.rng)
        for input_type, (X, y) in self.test_data.items():
            if input_type[-2:] == '1D':
                shift_std = np.array([0.1])
            else:
                shift_std = np.linspace(0.1, 1, 5)
            with self.subTest(transformer=generator.__class__, input_type=input_type):
                Xt, _ = generator.generate(X.copy(), y, shift_std=shift_std)
                # assert arrays have same size
                self.assertEqual(len(X), len(Xt))



import abc
from typing import Tuple

import numpy as np
from numpy.random import default_rng

from badgers.core.base import GeneratorMixin
from badgers.core.decorators.time_series import preprocess_inputs


class MissingValuesGenerator(GeneratorMixin):
    """
    Base class for transformers that generate point outliers in time-series data
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """
        :param random_generator: a random number generator
        :param n_outliers: the number of outliers to generate
        """
        self.random_generator = random_generator
        self.missing_indices_ = []

    @abc.abstractmethod
    def generate(self, X, y, **params) -> Tuple:
        pass


class MissingAtRandomGenerator(MissingValuesGenerator):
    """
    Randomly set data points to nan (missing)
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """

        :param random_generator: a random number generator

        """
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y, n_missing: int = 10) -> Tuple:
        """
        Randomly set values to np.nan (missing)
        :param X:
        :param y:
        :param n_missing: the number of outliers to generate
        :return:
        """
        # generate missing values indices and values
        rows = self.random_generator.choice(X.shape[0], size=n_missing, replace=False, p=None)
        cols = self.random_generator.integers(low=0, high=X.shape[1], size=n_missing)

        self.missing_indices_ = list(zip(rows, cols))

        for r, c in self.missing_indices_:
            X.iloc[r, c] = np.nan

        return X, y



from typing import Dict


class Pipeline:
    """
    Class for chaining different generators together.

    Not compatible with scikit-learn Pipelines
    Only words for python 3.7 as it relies upon dictionaries, and they need to be ordered
    """

    def __init__(self, generators: Dict):
        """
        Creates a Pipeline consisting of different generators
        :param generators: a dictionary containing as keys the generator names and as values the generators instances
        """
        self.generators = generators

    def generate(self, X, y, params: Dict = None):
        """
        Calls all generators generate function in the order

        :param X: the input features
        :param y: the class labels, regression targets, or None
        :param params: a dictionary containing as key: the generator names and
        as values: the parameters for each corresponding generate function
        :return: Xt, yt
        """
        if params is None:
            params = dict()
        for name, generator in self.generators.items():
            X, y = generator.generate(X, y, **params.get(name, dict()))
        return X, y



import unittest
from unittest import TestCase

import networkx as nx
import numpy as np

from badgers.generators.graph.missingness import NodesMissingCompletelyAtRandom, EdgesMissingCompletelyAtRandom


class TestNodesMissingCompletelyAtRandom(TestCase):
    def setUp(self) -> None:
        self.rng = np.random.default_rng(0)
        self.graph = nx.erdos_renyi_graph(n=100, p=0.25, seed=0, directed=False)

    def test_generate(self):
        percentage_missing = 0.1
        generator = NodesMissingCompletelyAtRandom(random_generator=self.rng)

        Xt, _ = generator.generate(X=self.graph, y=None, percentage_missing=percentage_missing)
        self.assertEqual(len(Xt), len(self.graph) - 10)

        Xt, yt = generator.generate(self.graph, [0] * len(self.graph))
        self.assertEqual(len(Xt), len(self.graph) - 10)
        self.assertEqual(len(yt), len(self.graph) - 10)


class TestEdgesMissingCompletelyAtRandom(TestCase):
    def setUp(self) -> None:
        self.rng = np.random.default_rng(0)
        self.graph = nx.erdos_renyi_graph(n=100, p=0.25, seed=0, directed=False)

    def test_generate(self):
        percentage_missing = 0.1
        generator = EdgesMissingCompletelyAtRandom(random_generator=self.rng)

        Xt, _ = generator.generate(X=self.graph, y=None, percentage_missing=percentage_missing)
        self.assertEqual(Xt.number_of_nodes(), self.graph.number_of_nodes())
        self.assertEqual(Xt.number_of_edges(),
                         len(self.graph.edges()) - int(self.graph.number_of_edges() * percentage_missing))

if __name__ == '__main__':
    unittest.main()


import abc
from copy import copy
from typing import Tuple

import networkx as nx
import numpy
import numpy as np
from numpy.random import default_rng

from badgers.core.base import GeneratorMixin


class MissingGenerator(GeneratorMixin):
    """
    Base class for missing nodes transformer
    """

    def __init__(self, random_generator: numpy.random.Generator = default_rng(seed=0)):
        """

        :param random_generator: A random generator
        """
        self.random_generator = random_generator

    @abc.abstractmethod
    def generate(self, X, y=None, **params) -> Tuple:
        pass


class NodesMissingCompletelyAtRandom(MissingGenerator):
    """
    Removes nodes from the graph uniformly at random.
    """

    def __init__(self, random_generator: numpy.random.Generator = default_rng(seed=0)):
        super().__init__(random_generator=random_generator)

    def generate(self, X, y=None, percentage_missing: float = 0.1) -> Tuple:
        """

        :param X:
        :param y:
        :param percentage_missing: The percentage of missing nodes (float value between 0 and 1 excluded)
        :return:
        """
        assert 0 < percentage_missing < 1
        if not isinstance(X, nx.Graph):
            raise NotImplementedError('badgers does only support networkx.Graph objects for graphs')

        nodes_to_be_removed = self.random_generator.choice(
            X.nodes(),
            int(X.number_of_nodes() * percentage_missing),
            replace=False
        )

        Xt = X.copy()
        Xt.remove_nodes_from(nodes_to_be_removed)

        if y is not None:
            yt = np.delete(y, nodes_to_be_removed)
        else:
            yt = None

        return Xt, yt


class EdgesMissingCompletelyAtRandom(MissingGenerator):
    """
    Removes edges from the graph uniformly at random.
    """

    def __init__(self, random_generator: numpy.random.Generator = default_rng(seed=0)):
        super().__init__(random_generator=random_generator)

    def generate(self, X, y=None, percentage_missing: float = 0.1) -> Tuple:
        """

        :param X:
        :param y:
        :param percentage_missing: The percentage of missing nodes (float value between 0 and 1 excluded)
        :return:
        """
        assert 0 < percentage_missing < 1
        if not isinstance(X, nx.Graph):
            raise NotImplementedError('badgers does only support networkx.Graph objects for graphs')

        edges_to_be_removed = self.random_generator.choice(
            X.edges(),
            int(X.number_of_edges() * percentage_missing),
            replace=False
        )

        Xt = X.copy()
        Xt.remove_edges_from(edges_to_be_removed)

        if y is None:
            yt = None
        elif isinstance(y, dict):
            yt = copy(y)
            for e in edges_to_be_removed:
                del yt[e]
        else:
            raise ValueError(f'This type of y is not supported {type(y)}, {y}')

        return Xt, yt



import unittest

import numpy as np

from badgers.generators.time_series.missingness import MissingAtRandomGenerator


class TestMissingAtRandomGenerator(unittest.TestCase):

    def setUp(self) -> None:
        self.t = np.linspace(0, 4 * np.pi, 101)
        self.X = np.sin(self.t).reshape(-1, 1)

    def test_generator(self):
        n_missing = 10
        generator = MissingAtRandomGenerator()
        Xt, _ = generator.generate(X=self.X, y=None, n_missing=n_missing)
        self.assertEqual(Xt.shape, self.X.shape)
        self.assertEqual(len(generator.missing_indices_), n_missing)
        self.assertEqual(Xt.isna().sum()[0], n_missing)


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest import TestCase

import numpy as np
import pandas as pd
from numpy.random import default_rng
from sklearn.datasets import make_blobs
from sklearn.decomposition import PCA

from badgers.generators.tabular_data.outliers import ZScoreSamplingGenerator, HistogramSamplingGenerator, \
    HypersphereSamplingGenerator, LowDensitySamplingGenerator, DecompositionAndOutlierGenerator, \
    IndependentHistogramsGenerator
from tests.generators.tabular_data import generate_test_data_only_features


class TestOutliersGenerator(TestCase):
    """
    Implements generic tests for all OutliersGenerator objects
    """

    def assert_shape_yt(self, yt, outliers):
        """
        asserts that yt and outliers have the same length
        """
        self.assertEqual(len(yt), len(outliers))

    def assert_shape_outliers(self, X, outliers, n_outliers):
        """
        asserts that the correct number of outliers have been produced
        with the correct number of features
        """
        if type(X) is not pd.DataFrame:
            X = pd.DataFrame(X)
        self.assertEqual(outliers.shape[0], n_outliers)
        self.assertEqual(outliers.shape[1], X.shape[1])


class TestZScoreSamplingGenerator(TestOutliersGenerator):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.generator = ZScoreSamplingGenerator(random_generator=self.rng)
        self.input_test_data = generate_test_data_only_features(rng=self.rng)

    def assert_zscore_larger_than_3(self, X, outliers):
        """
        Asserts that, at least in one dimension, the zscore of the generated outliers data points is greater than 3
        """
        # compute means and stds for checking z-score
        means = np.mean(X, axis=0)
        stds = np.std(X, axis=0)

        # assert z-score > 3
        for row in range(outliers.shape[0]):
            values = outliers[row, :]
            z_scores = abs(values - means) / stds
            self.assertTrue(all(z_scores > 3.))

    def test_generator(self):
        """

        """
        n_outliers = 10
        for input_type, (X, y) in self.input_test_data.items():
            outliers, yt = self.generator.generate(X.copy(), y, n_outliers=n_outliers)
            with self.subTest(input_type=input_type):
                self.assert_zscore_larger_than_3(X, outliers)
                self.assert_shape_yt(yt, outliers)
                self.assert_shape_outliers(X, outliers, n_outliers=n_outliers)


class TestHistogramSamplingGenerator(TestOutliersGenerator):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.generator = HistogramSamplingGenerator(random_generator=self.rng)
        self.input_test_data = generate_test_data_only_features(rng=self.rng)

    def test_generator(self):
        """

        """
        n_outliers = 10
        bins = 3
        for input_type, (X, y) in self.input_test_data.items():
            if input_type[-2:] == '2D':
                with self.subTest(input_type=input_type, ncols=10):
                    with self.assertRaises(NotImplementedError):
                        _, _ = self.generator.generate(X.copy(), y, n_outliers=n_outliers, bins=bins)

                    with self.subTest(input_type=input_type, ncols=3):
                        X = pd.DataFrame(X).iloc[:,:3]
                        outliers, yt = self.generator.generate(X, y, n_outliers=n_outliers, bins=bins)
                        self.assert_shape_yt(yt, outliers)
                        self.assert_shape_outliers(X, outliers, n_outliers)


class TestHypersphereSamplingGenerator(TestOutliersGenerator):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.generator = HypersphereSamplingGenerator(random_generator=self.rng)
        self.input_test_data = generate_test_data_only_features(rng=self.rng)

    def test_generator(self):
        """

        """
        n_outliers = 10
        for input_type, (X, y) in self.input_test_data.items():
            outliers, yt = self.generator.generate(X=X.copy(), y=y, n_outliers=n_outliers)
            with self.subTest(input_type=input_type):
                self.assert_shape_yt(yt, outliers)
                self.assert_shape_outliers(X, outliers, n_outliers)


class TestLowDensitySamplingGenerator(TestOutliersGenerator):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.generator = LowDensitySamplingGenerator(random_generator=self.rng)
        self.input_test_data = generate_test_data_only_features(rng=self.rng)

    def test_generator(self):
        """

        """
        n_outliers = 10
        for input_type, (X, y) in self.input_test_data.items():
            if input_type in ['numpy_1D', 'pandas_1D']:
                self.skipTest("Not testing numpy_1D and pandas_1D")
            else:
                outliers, yt = self.generator.generate(X=X.copy(), y=y, n_outliers=n_outliers)


class TestIndependentHistogramsGenerator(TestOutliersGenerator):

    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.generator = IndependentHistogramsGenerator(random_generator=self.rng)
        self.input_test_data = generate_test_data_only_features(rng=self.rng)

    def test_generator(self):
        """

        """
        n_outliers = 10
        bins = 3
        for input_type, (X, y) in self.input_test_data.items():
            outliers, yt = self.generator.generate(X.copy(), y, n_outliers=n_outliers, bins=bins)
            with self.subTest(input_type=input_type):
                self.assert_shape_yt(yt, outliers)
                self.assert_shape_outliers(X, outliers, n_outliers)


class TestDecompositionAndOutlierGenerator(TestOutliersGenerator):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.input_test_data = generate_test_data_only_features(rng=self.rng)

    def test_generator(self):
        """

        """
        X, y = make_blobs(centers=5, n_features=10, n_samples=100)
        n_outliers = 10
        outliers_generators = [
            ZScoreSamplingGenerator(),
            HistogramSamplingGenerator(),
            HypersphereSamplingGenerator(),
            LowDensitySamplingGenerator()
        ]
        for outlier_generator in outliers_generators:
            generator = DecompositionAndOutlierGenerator(
                decomposition_transformer=PCA(n_components=3),
                outlier_generator=outlier_generator
            )
            outliers, yt = generator.generate(X.copy(), y, n_outliers=n_outliers)
            with self.subTest(outlier_generator=outlier_generator):
                self.assert_shape_yt(yt, outliers)
                self.assert_shape_outliers(X, outliers, n_outliers)


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
import pandas as pd
from numpy.random import default_rng

from badgers.generators.time_series.changepoints import RandomChangeInMeanGenerator


class TestChangePointGenerator(unittest.TestCase):
    def setUp(self):
        self.random_generator = default_rng(seed=0)
        self.n_changepoints = 10
        self.min_change = -5
        self.max_change = 5

    def test_RandomChangeInMeanGenerator_generate(self):
        generator = RandomChangeInMeanGenerator(random_generator=self.random_generator)
        X = pd.DataFrame(data=np.zeros(100), columns=['dimension_0'], dtype=float)
        y = None

        Xt, _ = generator.generate(X.copy(), y, n_changepoints=self.n_changepoints, min_change=self.min_change,
                                   max_change=self.max_change)
        self.assertTrue(any(X != Xt))
        self.assertEqual(y, y)


if __name__ == '__main__':
    unittest.main()



from unittest import TestCase

import numpy as np

from badgers.generators.time_series.noise import GlobalGaussianNoiseGenerator, LocalGaussianNoiseGenerator, NoiseGenerator





class TestLocalGaussianNoiseGenerator(TestCase):

    def setUp(self) -> None:
        t = np.linspace(1, 10, 101)
        self.X = (np.sin(t * 2 * np.pi) + 0.5).reshape(-1, 1)

    def test_generate(self):
        generator = LocalGaussianNoiseGenerator()
        Xt, _ = generator.generate(self.X.reshape(-1, 1), None)


class TestGlobalGaussianNoiseGenerator(TestCase):

    def setUp(self) -> None:
        t = np.linspace(1, 10, 101)
        self.X = (np.sin(t * 2 * np.pi) + 0.5).reshape(-1, 1)

    def test_generate(self):
        generator = GlobalGaussianNoiseGenerator()
        Xt, _ = generator.generate(self.X.reshape(-1, 1), None)



import functools

import numpy as np
import pandas as pd


def preprocess_inputs(generate_func):
    """
    Validates and convert X and y for tabular data generators

    Preprocessing:
    X is converted to a pandas DataFrame
    y is converted to a pandas Series
    """

    @functools.wraps(generate_func)
    def wrapper(self, X, y, **kwargs):
        # Validate and preprocess X
        if isinstance(X, list):
            X = pd.DataFrame(X)
        elif isinstance(X, np.ndarray):
            # if it is a numpy array first check the dimensionality, if dimension is 1 then reshape, if the dimension > 2 then raise error
            if X.ndim == 1:
                X = X.reshape(-1, 1)
            if X.ndim > 2:
                raise ValueError(
                    "X has more than 2 dimensions where it is expected to have either 1 or 2!"
                )
            X = pd.DataFrame(data=X)
        elif isinstance(X, pd.Series):
            X = X.to_frame()
        elif isinstance(X, pd.DataFrame):
            # do nothing here
            pass
        else:
            raise ValueError(f"X must be a list, numpy array, pandas Series, or pandas DataFrame\nX is: {type(X)}")

        # Validate and preprocess y
        if y is not None:
            if isinstance(y, list):
                y = pd.Series(y)
            elif isinstance(y, np.ndarray):
                y = pd.Series(y)
            elif isinstance(y, pd.Series):
                pass
            else:
                raise ValueError("y must be a list, numpy array, or pandas Series")

        # Call the original function with the preprocessed inputs
        return generate_func(self, X, y, **kwargs)

    return wrapper



import unittest

import numpy as np

from badgers.generators.time_series.outliers import LocalZScoreGenerator, RandomZerosGenerator

class TestRandomZerosGenerator(unittest.TestCase):

    def setUp(self) -> None:
        self.t = np.linspace(0, 4 * np.pi, 101)
        self.X = np.sin(self.t).reshape(-1, 1)

    def test_generator(self):
        n_outliers = 10
        generator = RandomZerosGenerator()
        Xt, _ = generator.generate(X=self.X, y=None, n_outliers=n_outliers)
        self.assertEqual(Xt.shape, self.X.shape)
        self.assertEqual(len(generator.outliers_indices_), n_outliers)

class TestLocalZScoreGenerator(unittest.TestCase):

    def setUp(self) -> None:
        self.t = np.linspace(0, 4 * np.pi, 101)
        self.X = np.sin(self.t).reshape(-1, 1)

    def test_generator(self):
        n_outliers = 10
        generator = LocalZScoreGenerator()
        Xt, _ = generator.generate(X=self.X, y=None, n_outliers=n_outliers)
        self.assertEqual(Xt.shape, self.X.shape)
        self.assertEqual(len(generator.outliers_indices_), n_outliers)
        self.assertFalse(Xt.isna().any()[0])



if __name__ == '__main__':
    unittest.main()



import abc
from typing import Tuple, Union

import numpy as np
import pandas as pd
from numpy.random import default_rng
from scipy.interpolate import CubicSpline

from badgers.core.base import GeneratorMixin
from badgers.core.decorators.time_series import preprocess_inputs
from badgers.generators.time_series.utils import generate_random_patterns_indices


def add_offset(values: np.array, offset: float = 0.) -> np.array:
    return values + offset


def add_linear_trend(values: np.array, start_value: float = 0., end_value: float = 1.) -> np.array:
    return values + np.linspace(start_value - values[0], end_value - values[-1], len(values))


def scale(values: np.array, scaling_factor: float = 1.) -> np.array:
    return values * scaling_factor


class Pattern:

    def __init__(self, values: np.array):
        """
        Pattern constructor
        :param values: a 1D or 2D numpy array (rows = Time axis, columns = Features), if a single feature is used (1D), the values is automatically reshaped using reshape(-1,1)
        """
        if values.ndim == 1:
            values = values.reshape(-1, 1)
        if values.ndim > 2:
            raise ValueError(
                "Values has more thant 2 dimensions where it is expected to have either 1 or 2!"
            )
        self.values = values
        self.interpolation_function = CubicSpline(np.linspace(0, 1, values.shape[0]), values, bc_type='natural')

    def resample(self, nb_point: int) -> np.array:
        return self.interpolation_function(np.linspace(0, 1, nb_point))


class PatternsGenerator(GeneratorMixin):
    """
    Base class for transformers that generate patterns in time-series data
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """
        :param random_generator: a random number generator
        :param n_patterns: the number of patterns to generate
        """
        self.random_generator = random_generator
        self.patterns_indices_ = []

    @abc.abstractmethod
    def generate(self, X, y, **params) -> Tuple:
        pass

    def _inject_pattern(self, X: pd.DataFrame, p: Pattern, start_index: int, end_index: int,
                        scaling_factor: Union[float,str] = 'auto'):
        """
        Utility function to inject a predefined pattern `p` into a signal `X`
        :param X: the signal to inject the pattern
        :param p: the pattern to be injected
        :param start_index:
        :param end_index:
        :param scaling_factor: float | None | "auto" (default "auto")
        :return: the transformed signal where the pattern has been injected
        """

        # start and end values
        v_start = X.iloc[start_index, :].values
        v_end = X.iloc[start_index, :].values

        # number of points needed for resampling
        nb_points = end_index - start_index + 1

        if scaling_factor == 'auto':
            # compute a scaling factor to make the pattern looks realistic
            scaling_factor = (np.max(X[start_index:end_index + 1]) - np.min(X[start_index:end_index + 1])) / (
                np.max(p.values) - np.min(p.values)) * self.random_generator.normal(1, 0.2)
        elif scaling_factor is None:
            # default to 1
            scaling_factor = 1.0

        transformed_pattern = p.resample(nb_point=nb_points)
        transformed_pattern = scale(transformed_pattern, scaling_factor=scaling_factor)
        transformed_pattern = add_linear_trend(start_value=v_start, end_value=v_end, values=transformed_pattern)

        X.iloc[start_index:end_index + 1, :] = transformed_pattern
        return X


class RandomlySpacedPatterns(PatternsGenerator):
    """
    Inject given patterns with random width and indices
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y, n_patterns: int = 10, min_width_pattern: int = 5,
                 max_width_patterns: int = 10,
                 pattern: Pattern = Pattern(values=np.array([0, 0, 0, 0, 0]))) -> Tuple:
        # generate patterns indices and values
        self.patterns_indices_ = generate_random_patterns_indices(
            random_generator=self.random_generator,
            n_patterns=n_patterns,
            signal_size=len(X),
            min_width_pattern=min_width_pattern,
            max_width_patterns=max_width_patterns)

        for (start, end) in self.patterns_indices_:
            X = self._inject_pattern(X, p=pattern, start_index=start, end_index=end, scaling_factor='auto')

        return X, y


class RandomlySpacedConstantPatterns(PatternsGenerator):
    """
    Generates constant patterns of constant value with random width and indices
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y, n_patterns: int = 10, min_width_pattern: int = 5,
                 max_width_patterns: int = 10,
                 constant_value: float = 0) -> Tuple:
        # generate patterns indices and values
        self.patterns_indices_ = generate_random_patterns_indices(
            random_generator=self.random_generator,
            n_patterns=n_patterns,
            signal_size=len(X),
            min_width_pattern=min_width_pattern,
            max_width_patterns=max_width_patterns)

        for (start, end) in self.patterns_indices_:
            X.iloc[start:end, :] = constant_value

        return X, y


class RandomlySpacedLinearPatterns(PatternsGenerator):
    """
    Generates patterns of constant slope (linear) with random width and indices
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y, n_patterns: int = 10, min_width_pattern: int = 5,
                 max_width_patterns: int = 10) -> Tuple:
        # generate patterns indices and values
        self.patterns_indices_ = generate_random_patterns_indices(
            random_generator=self.random_generator,
            n_patterns=n_patterns,
            signal_size=len(X),
            min_width_pattern=min_width_pattern,
            max_width_patterns=max_width_patterns)

        for (start, end) in self.patterns_indices_:
            for col in range(X.shape[1]):
                X.iloc[start:end, col] = np.linspace(X.iloc[start, col], X.iloc[end, col], end - start)

        return X, y



import abc
from typing import Tuple


class GeneratorMixin:

    @abc.abstractmethod
    def generate(self, X, y, **params) -> Tuple:
        """

        :param X: the input
        :param y: the target
        :param params: optional parameters
        :return: Xt, yt
        """
        pass




from typing import Tuple

import numpy as np
import numpy.random
from numpy.random import default_rng


def normalize_proba(p: np.array) -> np.array:
    """
    Make sure the probability array respects the following constraints:
     - the sum of each column must be equal to 1 (or very close to 1)

    :param p: The probability array to normalize
    :return: The normalized array
    """
    # make the sum of each column = 1
    sum = p.sum(axis=0)
    # assure that no division by 0
    if isinstance(sum, np.ndarray) and sum.ndim > 1:
        sum[sum == 0] = 1
    # normalize
    p = p / sum
    return p


def random_sign(random_generator: numpy.random.Generator = default_rng(0), size: Tuple[int] = (1,)) -> np.array:
    """
    Generates an array full of ones, with randomly assigned signs.

    :param random_generator: a random number generator
    :param size: the shape of the array to generate
    :return: an array full of ones, with randomly assigned signs
    """
    signs = np.ones(shape=size)
    mask = random_generator.random(size=size) < 0.5
    signs[mask] *= -1
    return signs


def random_spherical_coordinate(random_generator: numpy.random.Generator = default_rng(0),
                                size: int = None,
                                radius: float = None) -> np.array:
    """
    Randomly generates points on a hypersphere of dimension `size`
    :param random_generator: a random number generator
    :param size: the dimension of the hypersphere
    :param radius: the radius of the hypersphere
    :return: an array of shape (`size`,) containing the values of the point generated
    """
    assert size > 0
    if size == 1:
        x = random_sign(random_generator, size=(1,)) * radius
    elif size == 2:
        phi = random_generator.uniform(0, 2. * np.pi)
        x = np.array([radius * np.cos(phi), radius * np.sin(phi)])
    else:
        phis = np.concatenate([
            random_generator.uniform(0, np.pi, size=size - 2),
            [random_generator.uniform(0, 2. * np.pi)]
        ])

        cos_phis = np.cos(phis)
        sin_phis = np.sin(phis)

        x = np.array(
            [radius * cos_phis[i] * np.prod(sin_phis[:i]) for i in range(size - 1)] + [radius * np.prod(sin_phis)]
        )
    return x



import abc

import numpy as np
import numpy.random
from numpy.random import default_rng

from badgers.core.base import GeneratorMixin
from badgers.core.decorators.tabular_data import preprocess_inputs
from badgers.core.utils import normalize_proba


class MissingValueGenerator(GeneratorMixin):
    """
    Base class for missing values transformer
    """

    def __init__(self, random_generator: numpy.random.Generator = default_rng(seed=0)):
        """
        :param random_generator: A random generator
        """
        self.random_generator = random_generator
        self.missing_values_indices_ = None

    @abc.abstractmethod
    def generate(self, X, y, **params):
        pass


class MissingCompletelyAtRandom(MissingValueGenerator):
    """
    A generator that removes values completely at random (MCAR [1]) (uniform distribution over all data).

    See also [1] https://stefvanbuuren.name/fimd/sec-MCAR.html
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """
        :param random_generator: A random generator
        """
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y, percentage_missing: float = 0.1):
        """
        Computes indices of missing values using a uniform distribution.

        :param X: the input features
        :param y: the target
        :param percentage_missing: The percentage of missing values (float value between 0 and 1 included)
        :return: Xt, yt
        """
        assert 0 <= percentage_missing <= 1
        # compute number of missing values per column
        nb_missing = int(X.shape[0] * percentage_missing)
        # generate missing values indices
        self.missing_values_indices_ = []
        for col in range(X.shape[1]):
            rows = self.random_generator.choice(X.shape[0], size=nb_missing, replace=False, p=None)
            self.missing_values_indices_ += [(row, col) for row in rows]
            # generate missing values
            X.iloc[rows, col] = np.nan

        return X, y


class DummyMissingAtRandom(MissingValueGenerator):
    """
    A generator that removes values at random (MAR [1]),
    where the probability of a data instance X[_,i] missing depends upon another feature X[_,j],
    where j is randomly chosen.

    See also [1] https://stefvanbuuren.name/fimd/sec-MCAR.html
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """
        :param random_generator: A random generator
        """
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y, percentage_missing: float = 0.1):
        """

        :param X: the input features
        :param y: the target
        :param percentage_missing: The percentage of missing values (float value between 0 and 1 included)
        :return: Xt, yt
        """
        assert 0 <= percentage_missing <= 1
        # initialize probability with zeros
        p = np.zeros_like(X)
        # normalize values between 0 and 1
        X_norm = ((X.max(axis=0) - X) / (X.max(axis=0) - X.min(axis=0))).values
        # make columns i depends on all the other
        if X.shape[1] > 1:
            for i in range(X.shape[1]):
                j = self.random_generator.choice([x for x in range(X.shape[1]) if x != i])
                p[:, i] = X_norm[:, j]
        else:
            p = X_norm
        p = normalize_proba(p)

        # compute number of missing values per column
        nb_missing = int(X.shape[0] * percentage_missing)
        # generate missing values indices
        self.missing_values_indices_ = []
        for col in range(X.shape[1]):
            rows = self.random_generator.choice(X.shape[0], size=nb_missing, replace=False, p=p[:, col])
            self.missing_values_indices_ += [(row, col) for row in rows]
            # generate missing values
            X.iloc[rows, col] = np.nan

        return X, y


class DummyMissingNotAtRandom(MissingValueGenerator):
    """
    A generator that removes values not at random (MNAR [1]),
    where the probability of a data instance X[i,j] missing depends linearly upon its own value.
    A data point X[i,j] = max(X[:,j]) has a missing probability of 1.
    A data point X[i,j] = min(X[:,j]) has a missing probability of 0.
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """
        :param random_generator: A random generator
        """
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y, percentage_missing):
        """

        :param X: the input features
        :param y: the target
        :param percentage_missing: The percentage of missing values (float value between 0 and 1 included)
        :return: Xt, yt
        """
        assert 0 <= percentage_missing <= 1

        # normalize values between 0 and 1
        p = ((X.max(axis=0) - X) / (X.max(axis=0) - X.min(axis=0))).values
        # make the sum of each column = 1
        p = normalize_proba(p)

        # compute number of missing values per column
        nb_missing = int(X.shape[0] * percentage_missing)
        # generate missing values indices
        self.missing_values_indices_ = []
        for col in range(X.shape[1]):
            rows = self.random_generator.choice(X.shape[0], size=nb_missing, replace=False, p=p[:, col])
            self.missing_values_indices_ += [(row, col) for row in rows]
            # generate missing values
            X.iloc[rows, col] = np.nan

        return X, y



import numpy as np
import numpy.random
import pandas as pd
from numpy.random import default_rng


def generate_test_data_only_features(rng: numpy.random.Generator = default_rng(0)):
    """

    :param rng:
    :return:
    """
    return {
        'list_1D': (rng.normal(size=100).tolist(), None),
        'list_2D': (rng.normal(size=(100, 10)).tolist(), None),
        'numpy_1D': (rng.normal(size=100).reshape(-1, 1), None),
        'numpy_2D': (rng.normal(size=(100, 10)), None),
        'pandas_1D': (
            pd.DataFrame(
                data=rng.normal(size=100).reshape(-1, 1),
                columns=['col']
            ), None),
        'pandas_2D': (
            pd.DataFrame(
                data=rng.normal(size=(100, 10)),
                columns=[f'col{i}' for i in range(10)]
            ), None),
    }


def generate_test_data_with_classification_labels(rng: numpy.random.Generator = default_rng(0)):
    """

    :param rng:
    :return:
    """
    return {
        'list_1D': (rng.normal(size=100).reshape(-1, 1).tolist(), [i % 5 for i in range(100)]),
        'list_2D': (rng.normal(size=(100, 10)).tolist(), [0, 1, 2, 3, 4] * 20),
        'numpy_1D': (rng.normal(size=100).reshape(-1, 1), np.array([i % 5 for i in range(100)])),
        'numpy_2D': (rng.normal(size=(100, 10)), np.array([0, 1, 2, 3, 4] * 20)),
        'pandas_1D': (
            pd.DataFrame(data=rng.normal(size=100).reshape(-1, 1),
                         columns=['col']),
            pd.Series(data=[i % 5 for i in range(100)])
        ),
        'pandas_2D': (
            pd.DataFrame(data=rng.normal(size=(100, 10)),
                         columns=[f'col{i}' for i in range(10)]),
            pd.Series(data=[0, 1, 2, 3, 4] * 20)
        )
    }


def generate_test_data_with_regression_targets(rng: numpy.random.Generator = default_rng(0)):
    """

    :param rng:
    :return:
    """
    return {
        'list_1D': (rng.normal(size=100).reshape(-1, 1).tolist(), rng.normal(size=100).tolist()),
        'list_2D': (rng.normal(size=(100, 10)).tolist(), rng.normal(size=100).tolist()),
        'numpy_1D': (rng.normal(size=100).reshape(-1, 1), rng.normal(size=100)),
        'numpy_2D': (rng.normal(size=(100, 10)), rng.normal(size=100)),
        'pandas_1D': (
            pd.DataFrame(data=rng.normal(size=100).reshape(-1, 1),
                         columns=['col']),
            pd.Series(data=rng.normal(size=100))
        ),
        'pandas_2D': (
            pd.DataFrame(data=rng.normal(size=(100, 10)),
                         columns=[f'col{i}' for i in range(10)]),
            pd.Series(rng.normal(size=100))
        )
    }



import abc

import numpy as np
import pandas as pd
from numpy.random import default_rng
from sklearn.preprocessing import StandardScaler

from badgers.core.base import GeneratorMixin
from badgers.core.decorators.tabular_data import preprocess_inputs


class NoiseGenerator(GeneratorMixin):
    """
    Base class for generators that add noise to tabular data
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """
        :param random_generator: A random generator
        """
        self.random_generator = random_generator

    @abc.abstractmethod
    def generate(self, X, y, **params):
        pass


class GaussianNoiseGenerator(NoiseGenerator):
    """
    A generator that adds Gaussian white noise to the tabular data
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """

        :param random_generator: A random generator

        """
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y, noise_std):
        """
        Adds Gaussian white noise to the data.
        The data is first standardized (each column has a mean = 0 and variance = 1).
        The noise is generated from a normal distribution with standard deviation = `noise_std`.
        The noise is added to the data.

        :param X: the input
        :param y: the target
        :param noise_std: The standard deviation of the noise to be added
        :return: Xt, yt
        """
        # standardize X
        scaler = StandardScaler()
        # fit, transform
        Xt = scaler.fit_transform(X)
        # add noise
        Xt += self.random_generator.normal(loc=0, scale=noise_std, size=Xt.shape)
        # inverse standardization
        Xt = scaler.inverse_transform(Xt)
        return pd.DataFrame(data=Xt, columns=X.columns, index=X.index), y


class GaussianNoiseClassesGenerator(NoiseGenerator):
    """
    A generator that adds Gaussian white noise to each class separately.
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """

        :param random_generator: A random generator
        """
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y, noise_std_per_class=dict()):
        """
        Add Gaussian white noise to the data.
        the data is first standardized (each column has a mean = 0 and variance = 1).
        The noise is generated from a normal distribution with standard deviation = `noise_std`.
        The noise is added to the data.

        :param X: the input
        :param y: the target
                :param noise_std_per_class: A dictionary giving the standard deviation of the noise to be added for each class
            key = class labels, values = noise std for this given class
        :return: Xt, yt
        """
        # standardize X
        scaler = StandardScaler()
        # fit, transform
        scaler.fit(X)
        Xt = scaler.transform(X)
        # add noise and repeat for each class

        tmp_Xt = []
        tmp_yt = []
        for label, noise_std in noise_std_per_class.items():
            data_class = np.array(Xt[y == label])
            noisy_data_class = data_class + self.random_generator.normal(loc=0, scale=noise_std, size=data_class.shape)
            labels = np.array([label] * data_class.shape[0])
            tmp_Xt.append(noisy_data_class.copy())
            tmp_yt.append(labels)

        Xt = np.concatenate(tmp_Xt, axis=0)
        yt = np.concatenate(tmp_yt, axis=0)
        # inverse standardization
        Xt = scaler.inverse_transform(Xt)
        return pd.DataFrame(data=Xt, columns=X.columns, index=X.index), pd.Series(yt)



import abc
import warnings

import numpy as np
import sklearn.base
from numpy.random import default_rng
from sklearn.decomposition import PCA
from sklearn.neighbors import KernelDensity
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler

from badgers.core.base import GeneratorMixin
from badgers.core.decorators.tabular_data import preprocess_inputs
from badgers.core.utils import random_sign, random_spherical_coordinate


class OutliersGenerator(GeneratorMixin):
    """
    Base class for transformers that add outliers to tabular data
    """

    def __init__(self, random_generator: np.random.Generator=default_rng(seed=0)):
        """
        :param random_generator: A random generator
        """
        self.random_generator = random_generator

    @abc.abstractmethod
    def generate(self, X, y=None, **params):
        pass


class ZScoreSamplingGenerator(OutliersGenerator):
    """
    Randomly generates outliers as data points with a z-score > 3.
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """

        :param random_generator: A random generator
        """
        super().__init__(random_generator)

    @preprocess_inputs
    def generate(self, X, y, n_outliers: int = 10):
        """
        Randomly generates outliers as data points with a z-score > 3.

        1. Standardize the input data (mean = 0, variance = 1)
        3. Generate outliers as follows:
            - the sign is randomly chosen
            - for each dimension: the value is equal to 3 + a random number following an exponential distribution function
            with default parameters (see https://numpy.org/doc/stable/reference/random/generated/numpy.random.Generator.exponential.html)
        4. Inverse the standardization transformation

        :param X: the input features
        :param y: the class labels, target values or None (if none yt
        :param n_outliers: The number of outliers to generate
        :return:
        """

        # standardize X
        scaler = StandardScaler()

        # fit, transform
        scaler.fit(X)
        Xt = scaler.transform(X)

        # generate outliers
        outliers = np.array([
            random_sign(self.random_generator, size=Xt.shape[1]) * (
                3. + self.random_generator.exponential(size=Xt.shape[1]))
            for _ in range(n_outliers)
        ])

        # in case we only have 1 outlier, reshape the array to match sklearn convention
        if outliers.shape[0] == 1:
            outliers = outliers.reshape(1, -1)

        # add "outliers" as labels for outliers
        yt = np.array(["outliers"] * len(outliers))

        return scaler.inverse_transform(outliers), yt


class HypersphereSamplingGenerator(OutliersGenerator):
    """
    Generates outliers by sampling points from a hypersphere with radius at least 3 sigma
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """

        :param random_generator: A random generator

        """
        super().__init__(random_generator)

    @preprocess_inputs
    def generate(self, X, y=None, n_outliers: int = 10):
        """
        Randomly generates outliers as data points with a z-score > 3.

        1. Standardize the input data (mean = 0, variance = 1)
        3. Generate outliers on a hypersphere (see https://en.wikipedia.org/wiki/N-sphere#Spherical_coordinates):
            - angles are chosen uniformly at random
            - radius is = 3 + a random number following an exponential distribution function with default parameters (see https://numpy.org/doc/stable/reference/random/generated/numpy.random.Generator.exponential.html)
        4. Inverse the standardization transformation

        :param X: the input features
        :param y: not used
        :param n_outliers: The number of outliers to generate
        :return:
        """

        # standardize X
        scaler = StandardScaler()

        # fit, transform
        scaler.fit(X)
        Xt = scaler.transform(X)

        # computing outliers
        outliers = np.array([
            random_spherical_coordinate(
                random_generator=self.random_generator,
                size=Xt.shape[1],
                radius=3. + self.random_generator.exponential()
            )
            for _ in range(n_outliers)
        ])

        # in case we only have 1 outlier, reshape the array to match sklearn convention
        if outliers.shape[0] == 1:
            outliers = outliers.reshape(1, -1)

        # add "outliers" as labels for outliers
        yt = np.array(["outliers"] * len(outliers))

        return scaler.inverse_transform(outliers), yt


class IndependentHistogramsGenerator(OutliersGenerator):
    """
    Randomly generates outliers from low density regions.
    Low density regions are estimated through several independent histograms (one for each feature).

    For each feature (column), a histogram is computed (it approximates the marginal distribution).
    Values are generated from bins with a low number of data points.

    All values generated for each feature are simply concatenated (independence hypothesis!).
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y=None, n_outliers: int = 10, bins: int = 10):
        """
        Randomly generates outliers from low density regions.
        Low density regions are estimated through several independent histograms (one for each feature).

        For each feature (column), a histogram is computed (it approximates the marginal distribution).
        Values are generated from bins with a low number of data points.

        All values generated for each feature are simply concatenated (independence hypothesis!).

        :param X: the input features
        :param y: not used
        :param params:
        :return:
        """
        outliers = []

        # loop over all features (columns)
        for col in range(X.shape[1]):
            # compute histogram of the current feature
            hist, bin_edges = np.histogram(X.iloc[:, col], bins=bins)
            # compute inverse density
            inv_density = 1 - hist / np.max(hist)
            # the sampling probability is proportional to the inverse density
            p = inv_density / np.sum(inv_density)
            # generate values:
            # first, choose randomly from which bin the value must be sampled
            indices = self.random_generator.choice(bins, p=p, size=n_outliers, replace=True)
            # second, sample uniformly at random from the selected bin
            values = [self.random_generator.uniform(low=bin_edges[i], high=bin_edges[i + 1]) for i in indices]
            # append the values for the current feature
            outliers.append(values)
        # cast as a numpy array
        outliers = np.array(outliers).T

        # add "outliers" as labels for outliers
        yt = np.array(["outliers"] * len(outliers))

        return outliers, yt


class HistogramSamplingGenerator(OutliersGenerator):
    """
    Randomly generates outliers from low density regions.
    Low density regions are estimated through a histogram.

    -----------------------------------------
    WARNING:
    This computes a full histogram in d-dimensions (d = nb features / columns), which is O(d²).
    Should only be used with low dimensionality data!
    It will raise an error if the number of dimensions is greater than 5.
    -----------------------------------------

    TODO: this works but is very inefficient, better strategies are welcome!
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """

        :param random_generator: A random generator

        """

        super().__init__(random_generator)


    @preprocess_inputs
    def generate(self, X, y=None, n_outliers: int = 10,
                 threshold_low_density: float = 0.1, bins: int = 10):
        """
        Randomly generates outliers from low density regions. Low density regions are estimated through histograms

        1. Standardize the input data (mean = 0, variance = 1)
        2. Compute and normalize histogram for the data
        3. Sample datapoint uniformly at random within bins of low density
        4. Inverse the standardization transformation

        :param X: the input features
        :param y: not used
        :param n_outliers: The number of outliers to generate
        :param threshold_low_density: the threshold that defines a low density region (must be between 0 and 1)
        :param bins: number of bins for the histogram
        :return:
        """
        assert 0 < threshold_low_density < 1
        if X.shape[1] > 5:
            raise NotImplementedError('So far this generator only supports tabular data with at most 5 columns')
        # standardize X
        scaler = StandardScaler()
        # fit, transform
        scaler.fit(X)
        Xt = scaler.transform(X)

        # compute the histogram of the data
        hist, edges = np.histogramdd(Xt, density=False, bins=bins)
        # normalize
        norm_hist = hist / (np.max(hist) - np.min(hist))
        # get coordinates of the histogram where the density is low (below a certain threshold)
        hist_coords_low_density = np.where(norm_hist <= threshold_low_density)
        # randomly pick some coordinates in the histogram where the density is low
        hist_coords_random = self.random_generator.choice(list(zip(*hist_coords_low_density)), n_outliers,
                                                          replace=True)

        # computing outliers values
        outliers = np.array([
            [
                self.random_generator.uniform(low=edges[i][c], high=edges[i][c + 1])
                for i, c in enumerate(h_coords)
            ]
            for h_coords in hist_coords_random
        ])

        # in case we only have 1 outlier, reshape the array to match sklearn convention
        if outliers.shape[0] == 1:
            outliers = outliers.reshape(1, -1)

        # add "outliers" as labels for outliers
        yt = np.array(["outliers"] * len(outliers))

        return scaler.inverse_transform(outliers), yt


class LowDensitySamplingGenerator(OutliersGenerator):
    """
    Randomly generates outliers from low density regions.
    Low density regions are estimated using a KernelDensity estimator.
    Points are sampled uniformly at random and filtered out if they do not belong to a low density region

    TODO: this works but might not be efficient, a better sampling strategy is welcome
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """

        :param random_generator: A random generator
        """
        super().__init__(random_generator=random_generator)
        self.density_estimator = KernelDensity(bandwidth="scott")


    @preprocess_inputs
    def generate(self, X, y=None, n_outliers: int = 10, threshold_low_density: float = 0.1, max_samples: int = 100):
        """
        Generate data points belonging to low density regions.

        Pseudo code:
        - Standardize the data X
        - Estimate the density based upon the original data X
        - Computes a threshold for determining low density (so far 10th percentile)
        - Sample uniformly at random within the hypercube [min, max]
        - Estimate the density of the new points and filter out the ones with a density that is above the threshold

        :param X: the input features
        :param y: not used
        :param n_outliers: The number of outliers to generate
        :param threshold_low_density: the threshold that defines a low density region (must be between 0 and 1)
        :param max_samples:
        :return:
        """
        assert 0 < threshold_low_density < 1
        # standardize X
        scaler = StandardScaler()
        # fit, transform
        scaler.fit(X)
        Xt = scaler.transform(X)
        # fit density estimator
        self.density_estimator = self.density_estimator.fit(Xt)
        low_density_threshold = np.percentile(self.density_estimator.score_samples(Xt), threshold_low_density)

        if max_samples is None:
            max_samples = n_outliers * 100

        outliers = np.array([
            x
            for x in self.random_generator.uniform(
                low=np.min(Xt, axis=0),
                high=np.max(Xt, axis=0),
                size=(max_samples, Xt.shape[1])
            )
            if self.density_estimator.score_samples(x.reshape(1, -1)) <= low_density_threshold
        ])

        if outliers.shape[0] < n_outliers:
            warnings.warn(
                f'LowDensitySamplingGenerator could not generate all {n_outliers} outliers. It only generated {len(outliers)}.')
        else:
            outliers = outliers[:n_outliers]

        # in case we only have 1 outlier, reshape the array to match sklearn convention
        if outliers.shape[0] == 1:
            outliers = outliers.reshape(1, -1)

        # add "outliers" as labels for outliers
        yt = np.array(["outliers"] * len(outliers))

        # in the case no outliers could be generated
        if outliers.shape[0] == 0:
            return outliers, yt

        return scaler.inverse_transform(outliers), yt


class DecompositionAndOutlierGenerator(OutliersGenerator):

    def __init__(self, decomposition_transformer: sklearn.base.TransformerMixin = PCA(n_components=2),
                 outlier_generator: OutliersGenerator = ZScoreSamplingGenerator(default_rng(0))):
        """

        :param decomposition_transformer: The dimensionality reduction transformer to be used before the outlier transformer
        :param outlier_generator: The outlier transformer to be used after the dimensionality has been reduced
        """
        assert hasattr(
            decomposition_transformer,
            'inverse_transform'), \
            f'the decomposition transformer class must implement the inverse_transform function.' \
            f'\nUnfortunately the class {decomposition_transformer} does not'
        super().__init__(random_generator=outlier_generator.random_generator)

        self.decomposition_transformer = decomposition_transformer
        self.outlier_generator = outlier_generator

    @preprocess_inputs
    def generate(self, X, y=None, **params):
        """
        Randomly generate outliers by first applying a dimensionality reduction technique (sklearn.decomposition)
        and an outlier transformer.

        1. Standardize the input data (mean = 0, variance = 1)
        2. Apply the dimensionality reduction transformer
        3. Generates outliers by applying the outlier transformer
        4. Inverse the dimensionality reduction and the standardization transformations

        :param X: the input features
        :param y: the regression target, class labels, or None
        :param params:
        :return:
        """

        # standardize the data and apply the dimensionality reduction transformer
        pipeline = make_pipeline(
            StandardScaler(),
            self.decomposition_transformer,
        )
        Xt = pipeline.fit_transform(X)
        # add outliers using the zscore_transformer
        Xt, yt = self.outlier_generator.generate(Xt, y, **params)
        # inverse the manifold and standardization transformations
        return pipeline.inverse_transform(Xt), yt



import unittest

import numpy as np
import pandas as pd
from numpy.random import default_rng
from pandas._testing import assert_frame_equal

from badgers.generators.time_series.trends import GlobalAdditiveLinearTrendGenerator


class TestTrendsGenerator(unittest.TestCase):
    def setUp(self):
        self.random_generator = default_rng(seed=0)

    def test_global_additive_linear_trend_generator(self):
        # Test the generate method of GlobalAdditiveLinearTrendGenerator
        global_additive_linear_trend_generator = GlobalAdditiveLinearTrendGenerator(
            random_generator=self.random_generator)
        X = pd.DataFrame(data=np.zeros(shape=(10, 4)), columns=[f'col{i}' for i in range(4)])
        y = None

        slope = np.array([1, 2, 3, 4])
        t = np.linspace(0, 1, len(X))
        expected_Xt = pd.DataFrame(data=t[:, np.newaxis] * slope, columns=X.columns, index=X.index)

        Xt, _ = global_additive_linear_trend_generator.generate(X, y, slope=slope)

        assert_frame_equal(Xt, expected_Xt)


if __name__ == '__main__':
    unittest.main()



from typing import List, Tuple

import numpy as np


def generate_random_patterns_indices(random_generator: np.random.Generator, signal_size: int, n_patterns: int,
                                     min_width_pattern: int, max_width_patterns: int) -> List[Tuple[int, int]]:
    """
    Generates a list of patterns indices (start index, stop index) randomly such that:
    - the patterns have a size in [min_width_pattern, max_width_patterns[
    - the patterns do not overlap
    - the patterns indices stays within the total signal size

    The algorithm is the following:

    1.  Compute patterns size (uniformly between min_width_pattern and max_width_patterns).

    2.  Split the total signal size into equal length segments.
        Each segment will contain exactly one pattern.
        So there are `n_patterns` segments.

    3.  For each segment, choose uniformly the starting and end points of the pattern such that the patterns fits into the segment.



    :param random_generator: a random generator
    :param signal_size: total signal size
    :param n_patterns: total number of patterns
    :param min_width_pattern: the minimal width of the patterns
    :param max_width_patterns: the maximal width of the patterns
    :return: a list of patterns indices (start index, stop index)
    """
    assert max_width_patterns * n_patterns < signal_size, 'the number of patterns * their maximum width should be smaller thant the total signal size'
    # randomly generates sizes for all patterns
    patterns_sizes = random_generator.integers(low=min_width_pattern, high=max_width_patterns, size=n_patterns)
    # size of segments (for splitting the total signal length)
    segment_size = signal_size // n_patterns
    # compute patterns indices
    patterns_indices = []
    for i in range(n_patterns - 1):
        start = random_generator.integers(low=i * segment_size, high=(i + 1) * segment_size - patterns_sizes[i])
        end = start + patterns_sizes[i]
        patterns_indices += [(start, end)]
    else:
        # take care of the last remaining segment
        start = random_generator.integers(low=(i + 1) * segment_size, high=signal_size - patterns_sizes[i + 1])
        end = start + patterns_sizes[i + 1]
        patterns_indices += [(start, end)]
    return patterns_indices



import abc
from typing import Union

import numpy as np
import pandas as pd
from numpy.random import default_rng
from sklearn.preprocessing import StandardScaler

from badgers.core.base import GeneratorMixin
from badgers.core.decorators.tabular_data import preprocess_inputs


class DriftGenerator(GeneratorMixin):
    """
    Base class for transformers that add noise to tabular data
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """
        :param random_generator: numpy.random.Generator, default default_rng(seed=0)
            A random generator
        """
        self.random_generator = random_generator

    @abc.abstractmethod
    def generate(self, X, y, **params):
        pass



class RandomShiftGenerator(DriftGenerator):
    """
    Randomly shift (geometrical translation) values of each column independently of one another.
    Data are first standardized (mean = 0, var = 1) and a random number is added to each column.
    The ith columns is simply translated: `$x_i \left arrow x_i + \epsilon_i$`
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """

        :param random_generator: A random generator
        :param shift_std: The standard deviation of the amount of shift applied (shift is chosen from a normal distribution)
        """
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y=None, shift_std: Union[float,np.array] = 0.1):
        """
        Randomly shift (geometrical translation) values of each column independently of one another.
        Data are first standardized (mean = 0, var = 1) and a random number is added to each column.
        The ith columns is simply translated: `$x_i \left arrow x_i + \epsilon_i$`


        :param X:
        :param y:
        :param shift_std:
        :return:
        """
        # normalize X
        scaler = StandardScaler()
        scaler.fit(X)
        Xt = scaler.transform(X)
        # generate random values for the shift for each column
        shift = self.random_generator.normal(loc=0, scale=shift_std, size=X.shape[1])
        # add shift
        Xt += shift
        # inverse transform
        return pd.DataFrame(data=scaler.inverse_transform(Xt), columns=X.columns, index=X.index), y


class RandomShiftClassesGenerator(DriftGenerator):
    """
    Randomly shift (geometrical translation) values of each class independently of one another.
    Data are first standardized (mean = 0, var = 1) and
    for each class a random number is added to all instances.
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """
        :param random_generator: A random generator
        """
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y, shift_std: Union[float,np.array] = 0.1):
        """
        Randomly shift (geometrical translation) values of each class independently of one another.
        Data are first standardized (mean = 0, var = 1) and
        for each class a random number is added to all instances.

        :param X:
        :param y:
        :param shift_std: The standard deviation of the amount of shift applied (shift is chosen from a normal distribution)
        """
        # extract unique labels
        classes = np.unique(y)
        # normalize X
        scaler = StandardScaler()
        scaler.fit(X)
        Xt = scaler.transform(X)
        # generate random values for the shift
        shifts = self.random_generator.normal(loc=0, scale=shift_std, size=len(classes))
        # add shift
        for c, s in zip(classes, shifts):
            Xt[y == c] += s
        # inverse transform
        return pd.DataFrame(data=scaler.inverse_transform(Xt), columns=X.columns, index=X.index), y



# Code adapted from: https://mkdocstrings.github.io/recipes/

from pathlib import Path
import mkdocs_gen_files

nav = mkdocs_gen_files.Nav()

src_dir = 'badgers'
reference_dir = 'reference'
for path in sorted(Path(src_dir).rglob('*.py')):
    module_path = path.with_suffix('')
    doc_path = path.with_suffix('.md')
    full_doc_path = Path(reference_dir, doc_path)

    parts = tuple(module_path.parts)
    if parts[-1] == '__init__':
        parts = parts[:-1]
        doc_path = doc_path.with_name('index.md')
        full_doc_path = full_doc_path.with_name('index.md')
    elif parts[-1] == '__main__':
        continue

    nav[parts] = doc_path.as_posix()

    with mkdocs_gen_files.open(full_doc_path, 'w') as fd:
        identifier = '.'.join(parts)
        print('::: ' + identifier, file=fd)

with mkdocs_gen_files.open(f'{reference_dir}/SUMMARY.md', 'w') as nav_file:
    nav_file.writelines(nav.build_literate_nav())



import abc
from typing import Tuple

from numpy.random import default_rng

from badgers.core.base import GeneratorMixin
from badgers.core.decorators.time_series import preprocess_inputs


class ChangePointsGenerator(GeneratorMixin):
    """
    Base class for generators that generate changepoints in time-series data
    """

    def __init__(self, random_generator=default_rng(seed=0), ):
        """
        :param random_generator: a random number generator
        """
        self.random_generator = random_generator
        self.changepoints = None

    @abc.abstractmethod
    def generate(self, X, y, **params) -> Tuple:
        pass


class RandomChangeInMeanGenerator(ChangePointsGenerator):
    """
    Generate randomly change in mean changepoints
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y, n_changepoints: int = 10, min_change: float = -5,
                 max_change: float = 5) -> Tuple:
        """

        :param X:
        :param y:
        :param max_change:
        :param min_change:
        :param n_changepoints:
        :return:
        """
        # Generate change points
        self.changepoints = list(
            zip(
                self.random_generator.integers(int(0.05 * len(X)), int(0.95 * len(X)), size=n_changepoints),
                self.random_generator.uniform(min_change, max_change, size=n_changepoints)
            )
        )

        for idx, change in self.changepoints:
            X.iloc[idx:] += change

        return X, y



import abc
from typing import Tuple

from numpy.random import default_rng

from badgers.core.base import GeneratorMixin


class TyposGenerator(GeneratorMixin):
    """
    Base class for transformers creating typos in a list of words
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """

        :param random_generator: numpy.random.Generator, default default_rng(seed=0)
            A random generator
        """
        self.random_generator = random_generator

    @abc.abstractmethod
    def generate(self, X, y, **params) -> Tuple:
        pass


class SwapLettersGenerator(TyposGenerator):
    """
    Swap adjacent letters in words randomly except for the first and the last letters.
    Example: 'kilogram' --> 'kilogarm'
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """

        :param random_generator: A random generator


        """
        super().__init__(random_generator)

    def generate(self, X, y, swap_proba:float=0.1) -> Tuple:
        """
        For each word with a length greater than 3, apply a single swap with probability `self.swap_proba`
        Where the swap happens is determined randomly


        :param X: A list of words where we apply typos
        :param y: not used
        :param swap_proba: Each word with a length greater than 3 will have this probability to contain a switch (max one per word)
        :return: the transformed list of words
        """
        for i in range(len(X)):
            if len(X[i]) > 3 and self.random_generator.random() <= swap_proba:
                # get the ith word in the list and make it a list of letters
                word = list(X[i])
                # randomly chose letters to switch
                idx = self.random_generator.integers(1, len(word) - 2)
                word[idx], word[idx + 1] = word[idx + 1], word[idx]
                # save the word with switched letters as string
                X[i] = ''.join(word)

        return X, y


class LeetSpeakGenerator(TyposGenerator):

    def __init__(self, random_generator=default_rng(seed=0)):
        """

        :param random_generator: a random number generator
        """
        super().__init__(random_generator=random_generator)
        self.leet_speak_mapping = {
            "A": ["4", "/\\", "@", "/-\\", "^", "(L", "\u0414"],
            "B": ["I3", "8", "13", "|3", "\u00df", "!3", "(3", "/3", ")3", "|-]", "j3"],
            "C": ["[", "\u00a2", "<", "(", "\u00a9"],
            "D": [")", "|)", "(|", "[)", "I>", "|>", "?", "T)", "I7", "cl", "|}", "|]"],
            "E": ["3", "&", "\u00a3", "\u20ac", "[-", "|=-"],
            "F": ["|=", "\u0192", "|#", "ph", "/=", "v"],
            "G": ["6", "&", "(_+", "9", "C-", "gee", "(?,", "[,", "{,", "<-", "(."],
            "H": ["#", "/-/", "\\-\\", "[-]", "]-[", ")-(", "(-)", ":-:", "|~|", "|-|", "]~[", "}{", "!-!", "1-1",
                  "\\-/", "I+I", "?"],
            "I": ["1", "|", "][", "!", "eye", "3y3"],
            "J": [",_|", "_|", "._|", "._]", "_]", ",_]", "]"],
            "K": [">|", "|<", "1<", "|c", "|(7<"],
            "L": ["1", "2", "\u00a3", "7", "|_", "|"],
            "M": ["/\\/\\", "/V\\", "[V]", "|\\/|", "^^", "<\\/>", "{V}", "(v)", "(V)", "|\\|\\", "]\\/[", "nn", "11"],
            "N": ["^/", "|\\|", "/\\/", "[\\]", "<\\>", "{\\}", "/V", "^", "\u0e17", "\u0418"],
            "O": ["0", "()", "oh", "[]", "p", "<>", "\u00d8"],
            "P": ["|*", "|o", "|\u00ba", "?", "|^", "|>", "|\"", "9", "[]D", "|\u00b0", "|7"],
            "Q": ["(_,)", "()_", "2", "0_", "<|", "&", "9", "\u00b6", "\u204b", "\u2117"],
            "R": ["I2", "9", "|`", "|~", "|?", "/2", "|^", "lz", "7", "2", "12", "\u00ae", "[z", "\u042f", ".-", "|2",
                  "|-", "3"],
            "S": ["5", "$", "z", "\u00a7", "ehs", "es", "2"],
            "T": ["7", "+", "-|-", "']['", "\u2020", "\u00ab|\u00bb", "~|~"],
            "U": ["(_)", "|_|", "v", "L|", "\u0e1a"],
            "V": ["\\/", "|/", "\\|"],
            "W": ["\\/\\/", "vv", "\\N", "'//", "\\\\'", "\\^/", "\\/\\/", "(n)", "\\V/", "\\X/", "\\|/", "\\_|_/",
                  "\\_:_/", "uu", "2u", "\\\\//\\\\//", "\u0e1e", "\u20a9"],
            "X": ["><", "}{", "ecks", "\u00d7", "?", "}{", ")(", "]["],
            "Y": ["j", "`/", "\\|/", "\u00a5", "\\//"],
            "Z": ["2", "7_", "-/_", "%", ">_", "s", "~/_", "-\\_", "-|_"]
        }

    def randomly_replace_letter(self, letter, replacement_proba):
        """
        Randomly replace a letter with its leet counterpart
        :param letter:
        :param replacement_proba: the probability of replacing a letter with its leet counterpart
        :return:
        """
        if self.random_generator.random() < replacement_proba:
            return self.random_generator.choice(self.leet_speak_mapping[letter.upper()])
        else:
            return letter

    def generate(self, X, y, replacement_proba: float = 0.1) -> Tuple:
        """

        :param X: A list of words where we apply leet replacement
        :param y:
        :param replacement_proba: the probability of replacing a letter with its leet counterpart
        :return:
        """
        assert 0 <= replacement_proba <= 1
        Xt = [
            ''.join([self.randomly_replace_letter(l, replacement_proba=replacement_proba) for l in word])
            for word in X
        ]

        return Xt, y



import abc
from typing import Tuple

import numpy as np
from numpy.random import default_rng

from badgers.core.base import GeneratorMixin
from badgers.core.decorators.time_series import preprocess_inputs
from badgers.core.utils import random_sign


class OutliersGenerator(GeneratorMixin):
    """
    Base class for transformers that generate point outliers in time-series data
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """
        :param random_generator: a random number generator
        :param n_outliers: the number of outliers to generate
        """
        self.random_generator = random_generator
        self.outliers_indices_ = []

    @abc.abstractmethod
    def generate(self, X, y, **params) -> Tuple:
        pass


class RandomZerosGenerator(OutliersGenerator):
    """
    Randomly set data points to 0
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """

        :param random_generator: a random number generator

        """
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y, n_outliers: int = 10) -> Tuple:
        """
        Randomly set values to zero
        :param X:
        :param y:
        :param n_outliers: the number of outliers to generate
        :return:
        """
        # generate extreme values indices and values
        rows = self.random_generator.choice(X.shape[0], size=n_outliers, replace=False, p=None)
        cols = self.random_generator.integers(low=0, high=X.shape[1], size=n_outliers)

        self.outliers_indices_ = list(zip(rows, cols))

        for r, c in self.outliers_indices_:
            X.iloc[r, c] = 0

        return X, y


class LocalZScoreGenerator(OutliersGenerator):
    """
    Randomly generates locally extreme values.
    Given a time interval (a window) of size l, an outlier is generated by setting the value a time t
    """

    def __init__(self, random_generator=default_rng(seed=0)):
        """

        :param random_generator: a random number generator
        """
        super().__init__(random_generator=random_generator)

    @preprocess_inputs
    def generate(self, X, y, n_outliers: int = 10,
                 local_window_size: int = 10):
        """
        Computes indices of extreme values using a uniform distribution.
        Computes the values at random outside the range [mean(X) - 3 sigma(X), mean(X) + 3 sigma(X)].
        The sign of the extreme value is the same as the value being replaced.

        :param X:
        :param y:
        :param n_outliers: the number of outliers to generate
        :param  local_window_size: the width (number of data points) of the local window to compute local Z-Score
        :return: the transformed array
        """
        # generate extreme values indices and values
        delta = int(local_window_size / 2)

        rows = self.random_generator.choice(
            np.arange(delta, X.shape[0] - delta, dtype=int), size=n_outliers, replace=False, p=None)
        cols = self.random_generator.integers(low=0, high=X.shape[1], size=n_outliers)

        self.outliers_indices_ = list(zip(rows, cols))

        for r, c in self.outliers_indices_:
            local_window = X.iloc[r - delta:r + delta, c]
            local_mean = local_window.mean(axis=0)
            local_std = local_window.std(axis=0)
            value = local_mean + random_sign(self.random_generator) * (
                3. * local_std + self.random_generator.exponential())
            # updating with new outliers
            X.iloc[r, c] = value

        return X, y



import unittest
from copy import deepcopy

from numpy.random import default_rng

from badgers.generators.text.typos import SwapLettersGenerator, LeetSpeakGenerator


class TestSwapLettersGenerator(unittest.TestCase):
    def test_transform(self):
        """
        For all words with a length strictly greater than 3:
        - The first and the last letter must be the same (i.e., these are not switched)
        - Both original and transformed words must have the same length
        - Both original and transformed words must have the same letters (i.e., letters are switched but not replaced)

        For smaller words (length lower than or equal to 3):
        - no change occurs
        """
        trf = SwapLettersGenerator(random_generator=default_rng(0))
        X = [
            'abcdef',
            'abcde',
            'abcd',
            'abc',
            'ab',
            'a'
        ]

        Xt, _ = trf.generate(X=deepcopy(X), y=None, swap_proba=1)

        for i in range(len(X)):
            if len(X[i]) > 3:
                self.assertNotEqual(Xt[i], X[i])
                self.assertEqual(Xt[i][0], X[i][0])
                self.assertEqual(Xt[i][-1], X[i][-1])
                self.assertEqual(len(Xt[i]), len(X[i]))
                self.assertEqual(set(Xt[i]), set(X[i]))
            else:
                self.assertEqual(Xt[i], X[i])


class TestLeetSpeakGenerator(unittest.TestCase):

    def test_generate(self):
        X = ['the', 'quick', 'brown', 'fox', 'jumps', 'over', 'the', 'lazy', 'fox']
        trf = LeetSpeakGenerator()
        Xt, _ = trf.generate(X, None)
        self.assertEqual(len(X), len(Xt))


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest import TestCase

import numpy as np
from numpy.random import default_rng

from badgers.core.utils import normalize_proba
from badgers.generators.tabular_data.imbalance import RandomSamplingFeaturesGenerator, \
    RandomSamplingClassesGenerator, RandomSamplingTargetsGenerator
from tests.generators.tabular_data import generate_test_data_with_classification_labels, \
    generate_test_data_with_regression_targets


class TestRandomSamplingClassesGenerator(TestCase):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.input_test_data = generate_test_data_with_classification_labels(rng=self.rng)

    def test_generate(self):

        for input_type, (X, y) in self.input_test_data.items():

            proportion_classes = {0: 0.5, 1: 0.2, 2: 0.1, 3: 0.1, 4: 0.1}

            generator = RandomSamplingClassesGenerator(random_generator=self.rng)

            Xt, yt = generator.generate(X.copy(), y, proportion_classes=proportion_classes)
            # assert arrays have same size
            if input_type[-2:] == '1D':
                self.assertEqual(Xt.shape[1], 1)
            else:
                self.assertEqual(Xt.shape[1], 10)
            self.assertEqual(Xt.shape[0], len(yt))
            # assert pandas DataFrame get the same columns before and after the call to generate function
            if input_type[:6] == 'pandas':
                self.assertListEqual(list(X.columns), list(Xt.columns))


class TestRandomSamplingFeaturesGenerator(TestCase):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.input_test_data = generate_test_data_with_classification_labels(rng=self.rng)

    def test_generate(self):
        def proba_func(X):
            feature = X.iloc[:, 0]
            return normalize_proba(
                (np.max(feature) - feature) / (np.max(feature) - np.min(feature))
            )

        generator = RandomSamplingFeaturesGenerator()

        for input_type, (X, y) in self.input_test_data.items():
            with self.subTest(transformer=generator.__class__, input_type=input_type):
                Xt, yt = generator.generate(X.copy(), y, sampling_proba_func=proba_func)
                # assert arrays have same size
                if input_type[-2:] == '1D':
                    self.assertEqual(Xt.shape[1], 1)
                else:
                    self.assertEqual(Xt.shape[1], 10)
                self.assertEqual(Xt.shape[0], len(yt))
                # assert pandas DataFrame get the same columns before and after the call to generate function
                if input_type in ['pandas_1D', 'pandas_2D']:
                    self.assertListEqual(list(X.columns), list(Xt.columns))


class TestRandomSamplingTargetsGenerator(TestCase):
    def setUp(self) -> None:
        self.rng = default_rng(0)
        self.input_test_data = generate_test_data_with_regression_targets(rng=self.rng)

    def test_generate(self):
        def proba_func(y):
            return normalize_proba(
                (np.max(y) - y) / (np.max(y) - np.min(y))
            )

        generator = RandomSamplingTargetsGenerator()

        for input_type, (X, y) in self.input_test_data.items():
            with self.subTest(transformer=generator.__class__, input_type=input_type):
                Xt, yt = generator.generate(X.copy(), y, sampling_proba_func=proba_func)
                # assert arrays have same size
                if input_type[-2:] == '1D':
                    self.assertEqual(Xt.shape[1], 1)
                else:
                    self.assertEqual(Xt.shape[1], 10)
                self.assertEqual(Xt.shape[0], len(yt))
                # assert pandas DataFrame get the same columns before and after the call to generate function
                if input_type in ['pandas_1D', 'pandas_2D']:
                    self.assertListEqual(list(X.columns), list(Xt.columns))


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
from numpy.random import default_rng

from badgers.generators.time_series.seasons import GlobalAdditiveSinusoidalSeasonGenerator


class TestSeasonsGenerator(unittest.TestCase):
    def setUp(self):
        self.random_generator = default_rng(seed=0)
        self.global_additive_season_generator = GlobalAdditiveSinusoidalSeasonGenerator(
            random_generator=self.random_generator)

    def test_global_additive_sinusoidal_season_generator(self):
        # Test the generate method of GlobalAdditiveSinusoidalSeasonGenerator
        X = np.zeros(shape=100)
        y = None
        period = 10
        generator = GlobalAdditiveSinusoidalSeasonGenerator(random_generator=self.random_generator)
        Xt, yt = generator.generate(X=X, y=y, period=period)

        t = np.arange(100)
        season = np.sin(t * 2 * np.pi / period)
        Xt_expected = X + season

        self.assertListEqual(Xt_expected.reshape(-1,1).tolist(), Xt.values.tolist())
        self.assertEqual(y, yt)


if __name__ == '__main__':
    unittest.main()
